utils::globalVariables(c("vigilance_fn_default", "vigilance_fn_default", "vigilance_fn_default"))
#' Specifies and trains a FuzzyART model
#'
#' @param inputs Specifies input data as a data.frame, matrix or simiar structure. Inputs should be normalized prior to running by using the \code{\link[=normalize]{normalize()}} function.
#' @param rho Vigilance parameter in (0,1).
#' @param alpha Choice parameter alpha > 0. Can be viewed as a regularization parameter penalizing large weights.
#' @param beta Learning rate in (0,1).
#' @param max_epochs Maximum number of iterations over the input.
#' @param shuffle TRUE allows the input to be shuffled during training. FALSE disables this.
#' @param random_seed A number specifying the seed for the random number generator while shuffling.
#' @param w_init Initial model weights.
#' @param distance_fn A distance function to determine how close an observation is to the category weight (see \code{\link[=fuzzy_and]{fuzzy_and()}} for specifications).
#' @param match_fn A match function like \code{\link[FuzzyART:match_fn_default]{match_fn_default()}}.
#' @param vigilance_fn A vigilance function like \code{\link[FuzzyART:vigilance_fn_default_one]{vigilance_fn_default_one()}}.
#' @param update_fn An update function like \code{\link[FuzzyART:update_fn_default]{update_fn_default()}}.
#' @param max_clusters The maximum numbers of cluster allowed.
#' @param eps Tolerance. Default is 10^-6.
#' @param show_status TRUE (default) will display the progress of learning measured as a percentage of max_epochs. FALSE will supress the status.
#' @param beta_decay A decay factor between 0 and 1. The closer to 0, the faster the decay.
#'
#' @return An object of class FuzzyART specifying the trained parameters, the labels for the input data, and the parameters used.
#' @export
#' @md
#'
#'
#' @references
#' Carpenter, Gail A., Stephen Grossberg, and David B. Rosen. "Fuzzy ART: Fast stable learning and categorization of analog patterns by an adaptive resonance system." Neural networks 4.6 (1991): 759-771.
#' @examples normalized_inputs = normalize(df = subset(iris,select = -Species))
#' mod  = FuzzyART_train(normalized_inputs,alpha = .8,rho = .5,beta = .08,
#'                       max_epochs = 1000,max_clusters =20,eps = 10^-8,random_seed = 4)
#' mod$Labels
#' plot(iris, col = mod$Labels)
FuzzyART_train = function(inputs, rho, alpha, beta, max_epochs, shuffle = TRUE,
                    random_seed = NA, w_init = NA, distance_fn = fuzzy_and,
                    match_fn = match_fn_default, vigilance_fn = vigilance_fn_default_one, update_fn = update_fn_default,
                    max_clusters = 1000, eps = 10^-6, show_status = TRUE, beta_decay = 1)
{
  ###### save parameters #########
  params = list()
  params[["rho"]]         = rho
  params[["alpha"]]       = alpha
  params[["beta"]]        = beta
  params[["max_epochs"]]  = max_epochs
  params[["shuffle"]]     = shuffle
  params[["random_seed"]] = random_seed
  params[["w_init"]]      = w_init
  params[["distance_fn"]] = distance_fn
  params[["match_fn"]]    = match_fn
  params[["update_fn"]]   = update_fn
  params[["update_fn"]]   = update_fn
  params[["vigilance_fn"]]= vigilance_fn
  params[["data"]]        = as.data.frame(inputs)
  ###### fit             ######
  inputs = normalize(inputs)
  num_features = dim(inputs)[2]
  if(!is.na(w_init))
  {
    # check if weight dimension matches input size
    stopifnot(dim(w_init)[2]==num_features*2)

    # initialiuw weights
    w = w_init
  }
  else{ w = matrix(rep(1,num_features*2),nrow = 1)}
  #colnames(w) <- rep(colnames(inputs), 2)
  num_clusters = dim(w)[1]
  dataset = cbind(inputs,1-inputs)

  #init variables
  num_clusters_old = num_clusters
  cluster_labels = rep(NA, dim(dataset)[1])
  iterations = 0
  w_old = w+1

  if(shuffle & !is.na(random_seed)) set.seed(random_seed)

  # learning until max_epochs reached or convergence
  train_output_lst = list(w = NA, num_clusters=NA, winner = NA)
  indices = 1:dim(dataset)[1]
  while((sum((w_old-w)^2)>eps^2) & (iterations< max_epochs))
  {
    if(show_status & iterations %% 100 ==0) print(paste0("Training FuzzyART in progress: ", round(iterations/max_epochs*100,digits = 2),"%."))

    w_old = w
    if(shuffle) indices <- sample(indices)

    for(i in indices)
    {
      train_output_lst <- train_pattern(dataset[i,],w,alpha,rho,beta,num_clusters,max_cluster_count = max_clusters,
                                        match_fn=match_fn, update_fn = update_fn,
                                        distance_fn = distance_fn, vigilance_fn = vigilance_fn,
                                        out.lst = train_output_lst)
      cluster_labels[i] <- train_output_lst$winner
      w <- train_output_lst$w
      #w_old <- train_output_lst$w_old
      num_clusters <- train_output_lst$num_clusters
    }
    iterations = iterations+1
    beta = beta*beta_decay
    #w and w_old conformable?
    if(num_clusters_old != num_clusters)
    {
      w_old = rbind(w_old,replicate(n = num_features*2, rep(1,num_clusters-num_clusters_old)))
      num_clusters_old = num_clusters
    }
  }


  FuzzyArt_Module = list()
  FuzzyArt_Module[["Labels"]]=factor(cluster_labels)
  FuzzyArt_Module[["weights"]]= w
  FuzzyArt_Module[["params"]] = params
  FuzzyArt_Module[["iterations"]] = iterations


  class(FuzzyArt_Module)<- "FuzzyART"
  return(FuzzyArt_Module)
}

#' Predicts the cluster labels for given input patterns based on a trained FuzzyART model. Alternatively use the predict function.
#'
#' @param inputs Specifies the input patterns.
#' @param FuzzyArt_Module A trained FuzzyArt_Module as created by FuzzyART_train.
#'
#' @return The prediced labels for the imput patterns.
#' @export
#'
#' @examples normalized_inputs = normalize(df = subset(iris,select = -Species))
#' mod  = FuzzyART_train(normalized_inputs,alpha = 0.3,rho = .6,beta = .02,
#'                       max_epochs = 100,max_clusters = 10)
#' FuzzyART_predict(normalized_inputs, mod)
FuzzyART_predict = function(inputs, FuzzyArt_Module)
{
  if(is.null(FuzzyArt_Module)) stop("Error. Specify FuzzyArt_Module as a FuzzyART model as trained by FuzzyART_train")
  dataset = cbind(inputs,inputs-1)
  labels = apply(dataset, MARGIN = 1,FUN = function(pattern){return(eval_pattern(pattern = pattern,w = FuzzyArt_Module$w,
                                                                      alpha = FuzzyArt_Module$params$alpha, rho = FuzzyArt_Module$params$rho,
                                                                      match_fn = FuzzyArt_Module$params$match_fn,
                                                                      vigilance_fn = FuzzyArt_Module$params$vigilance_fn))})
  return(as.factor(labels))
}



#' Normalizes the data to lie within the d dimensional hypercube $\[0,1\]^d$, where d is the number of features or columns of the data. This procedure should not be confused with standardization where the mean is centered at zero and the standard deviation is set to 1 (a.k.a. z-score).
#'
#' @param df a data.frame like structure to be normalized
#'
#' @return The normalized data.frame or data matrix.
#' @export
#'
#' @examples df = data.frame(a = c(1,-1,2), b = c(.1,-2,2))
#' normalize(df)
normalize = function(df)
{
  return(apply(X = df,MARGIN = 2,FUN = function(col){(col-min(col))/(max(col)-min(col))}))
}
